@@ -1,18 +1,6 @@
 from abc import ABCMeta, abstractmethod
-from . import _core
-
-__all__ = [
-    "Clock",
-    "Instrument",
-    "AsyncResource",
-    "SendStream",
-    "ReceiveStream",
-    "Stream",
-    "HalfCloseableStream",
-    "SocketFactory",
-    "HostnameResolver",
-    "Listener",
-]
+from typing import Generic, TypeVar
+import trio
 
 
 # We use ABCMeta instead of ABC, plus set __slots__=(), so as not to force a
@@ -91,7 +79,7 @@ class Instrument(metaclass=ABCMeta):
         """Called when the given task is created.
 
         Args:
-            task (trio.hazmat.Task): The new task.
+            task (trio.lowlevel.Task): The new task.
 
         """
 
@@ -102,7 +90,7 @@ class Instrument(metaclass=ABCMeta):
         runnable tasks ahead of it.
 
         Args:
-            task (trio.hazmat.Task): The task that became runnable.
+            task (trio.lowlevel.Task): The task that became runnable.
 
         """
 
@@ -110,7 +98,7 @@ class Instrument(metaclass=ABCMeta):
         """Called immediately before we resume running the given task.
 
         Args:
-            task (trio.hazmat.Task): The task that is about to run.
+            task (trio.lowlevel.Task): The task that is about to run.
 
         """
 
@@ -118,7 +106,7 @@ class Instrument(metaclass=ABCMeta):
         """Called when we return to the main run loop after a task has yielded.
 
         Args:
-            task (trio.hazmat.Task): The task that just ran.
+            task (trio.lowlevel.Task): The task that just ran.
 
         """
 
@@ -126,7 +114,7 @@ class Instrument(metaclass=ABCMeta):
         """Called when the given task exits.
 
         Args:
-            task (trio.hazmat.Task): The finished task.
+            task (trio.lowlevel.Task): The finished task.
 
         """
 
@@ -151,7 +139,7 @@ class Instrument(metaclass=ABCMeta):
 
 class HostnameResolver(metaclass=ABCMeta):
     """If you have a custom hostname resolver, then implementing
-    :class:`HostnameResolver` allows you to register this to be used by trio.
+    :class:`HostnameResolver` allows you to register this to be used by Trio.
 
     See :func:`trio.socket.set_custom_hostname_resolver`.
 
@@ -187,26 +175,25 @@ class HostnameResolver(metaclass=ABCMeta):
 
 
 class SocketFactory(metaclass=ABCMeta):
-    """If you write a custom class implementing the trio socket interface,
-    then you can use a :class:`SocketFactory` to get trio to use it.
+    """If you write a custom class implementing the Trio socket interface,
+    then you can use a :class:`SocketFactory` to get Trio to use it.
 
     See :func:`trio.socket.set_custom_socket_factory`.
 
     """
-
     @abstractmethod
     def socket(self, family=None, type=None, proto=None):
         """Create and return a socket object.
 
         Your socket object must inherit from :class:`trio.socket.SocketType`,
         which is an empty class whose only purpose is to "mark" which classes
-        should be considered valid trio sockets.
+        should be considered valid Trio sockets.
 
         Called by :func:`trio.socket.socket`.
 
         Note that unlike :func:`trio.socket.socket`, this does not take a
         ``fileno=`` argument. If a ``fileno=`` is specified, then
-        :func:`trio.socket.socket` returns a regular trio socket object
+        :func:`trio.socket.socket` returns a regular Trio socket object
         instead of calling this method.
 
         """
@@ -253,7 +240,7 @@ class AsyncResource(metaclass=ABCMeta):
         connection. This requires sending a "goodbye" message; but if the peer
         has become non-responsive, then our attempt to send this message might
         block forever, and eventually time out and be cancelled. In this case
-        the :meth:`aclose` method on :class:`~trio.ssl.SSLStream` will
+        the :meth:`aclose` method on :class:`~trio.SSLStream` will
         immediately close the underlying transport stream using
         :func:`trio.aclose_forcefully` before raising :exc:`~trio.Cancelled`.
 
@@ -282,8 +269,12 @@ class SendStream(AsyncResource):
     bidirectional, then you probably want to also implement
     :class:`ReceiveStream`, which makes your object a :class:`Stream`.
 
-    Every :class:`SendStream` also implements the :class:`AsyncResource`
-    interface.
+    :class:`SendStream` objects also implement the :class:`AsyncResource`
+    interface, so they can be closed by calling :meth:`~AsyncResource.aclose`
+    or using an ``async with`` block.
+
+    If you want to send Python objects rather than raw bytes, see
+    :class:`SendChannel`.
 
     """
     __slots__ = ()
@@ -296,16 +287,16 @@ class SendStream(AsyncResource):
           data (bytes, bytearray, or memoryview): The data to send.
 
         Raises:
-          trio.ResourceBusyError: if another task is already executing a
+          trio.BusyResourceError: if another task is already executing a
               :meth:`send_all`, :meth:`wait_send_all_might_not_block`, or
               :meth:`HalfCloseableStream.send_eof` on this stream.
-          trio.BrokenStreamError: if something has gone wrong, and the stream
+          trio.BrokenResourceError: if something has gone wrong, and the stream
               is broken.
           trio.ClosedResourceError: if you previously closed this stream
               object, or if another task closes this stream object while
               :meth:`send_all` is running.
 
-        Most low-level operations in trio provide a guarantee: if they raise
+        Most low-level operations in Trio provide a guarantee: if they raise
         :exc:`trio.Cancelled`, this means that they had no effect, so the
         system remains in a known state. This is **not true** for
         :meth:`send_all`. If this operation raises :exc:`trio.Cancelled` (or
@@ -329,10 +320,10 @@ class SendStream(AsyncResource):
         return. When implementing it, err on the side of returning early.
 
         Raises:
-          trio.ResourceBusyError: if another task is already executing a
+          trio.BusyResourceError: if another task is already executing a
               :meth:`send_all`, :meth:`wait_send_all_might_not_block`, or
               :meth:`HalfCloseableStream.send_eof` on this stream.
-          trio.BrokenStreamError: if something has gone wrong, and the stream
+          trio.BrokenResourceError: if something has gone wrong, and the stream
               is broken.
           trio.ClosedResourceError: if you previously closed this stream
               object, or if another task closes this stream object while
@@ -378,37 +369,43 @@ class ReceiveStream(AsyncResource):
     bidirectional, then you probably want to also implement
     :class:`SendStream`, which makes your object a :class:`Stream`.
 
-    Every :class:`ReceiveStream` also implements the :class:`AsyncResource`
-    interface.
+    :class:`ReceiveStream` objects also implement the :class:`AsyncResource`
+    interface, so they can be closed by calling :meth:`~AsyncResource.aclose`
+    or using an ``async with`` block.
+
+    If you want to receive Python objects rather than raw bytes, see
+    :class:`ReceiveChannel`.
+
+    `ReceiveStream` objects can be used in ``async for`` loops. Each iteration
+    will produce an arbitrary sized chunk of bytes, like calling
+    `receive_some` with no arguments. Every chunk will contain at least one
+    byte, and the loop automatically exits when reaching end-of-file.
 
     """
     __slots__ = ()
 
     @abstractmethod
-    async def receive_some(self, max_bytes):
+    async def receive_some(self, max_bytes=None):
         """Wait until there is data available on this stream, and then return
-        at most ``max_bytes`` of it.
+        some of it.
 
         A return value of ``b""`` (an empty bytestring) indicates that the
         stream has reached end-of-file. Implementations should be careful that
         they return ``b""`` if, and only if, the stream has reached
         end-of-file!
 
-        This method will return as soon as any data is available, so it may
-        return fewer than ``max_bytes`` of data. But it will never return
-        more.
-
         Args:
           max_bytes (int): The maximum number of bytes to return. Must be
-              greater than zero.
+              greater than zero. Optional; if omitted, then the stream object
+              is free to pick a reasonable default.
 
         Returns:
           bytes or bytearray: The data received.
 
         Raises:
-          trio.ResourceBusyError: if two tasks attempt to call
+          trio.BusyResourceError: if two tasks attempt to call
               :meth:`receive_some` on the same stream at the same time.
-          trio.BrokenStreamError: if something has gone wrong, and the stream
+          trio.BrokenResourceError: if something has gone wrong, and the stream
               is broken.
           trio.ClosedResourceError: if you previously closed this stream
               object, or if another task closes this stream object while
@@ -416,6 +413,15 @@ class ReceiveStream(AsyncResource):
 
         """
 
+    def __aiter__(self):
+        return self
+
+    async def __anext__(self):
+        data = await self.receive_some()
+        if not data:
+            raise StopAsyncIteration
+        return data
+
 
 class Stream(SendStream, ReceiveStream):
     """A standard interface for interacting with bidirectional byte streams.
@@ -459,7 +465,7 @@ class HalfCloseableStream(Stream):
           page <https://linux.die.net/man/2/shutdown>`__).
 
         * The SSH protocol provides the ability to multiplex bidirectional
-          "channels" on top of a single encrypted connection. A trio
+          "channels" on top of a single encrypted connection. A Trio
           implementation of SSH could expose these channels as
           :class:`HalfCloseableStream` objects, and calling :meth:`send_eof`
           would send an ``SSH_MSG_CHANNEL_EOF`` request (see `RFC 4254 §5.3
@@ -467,18 +473,18 @@ class HalfCloseableStream(Stream):
 
         * On an SSL/TLS-encrypted connection, the protocol doesn't provide any
           way to do a unidirectional shutdown without closing the connection
-          entirely, so :class:`~trio.ssl.SSLStream` implements
+          entirely, so :class:`~trio.SSLStream` implements
           :class:`Stream`, not :class:`HalfCloseableStream`.
 
         If an EOF has already been sent, then this method should silently
         succeed.
 
         Raises:
-          trio.ResourceBusyError: if another task is already executing a
+          trio.BusyResourceError: if another task is already executing a
               :meth:`~SendStream.send_all`,
               :meth:`~SendStream.wait_send_all_might_not_block`, or
               :meth:`send_eof` on this stream.
-          trio.BrokenStreamError: if something has gone wrong, and the stream
+          trio.BrokenResourceError: if something has gone wrong, and the stream
               is broken.
           trio.ClosedResourceError: if you previously closed this stream
               object, or if another task closes this stream object while
@@ -487,9 +493,31 @@ class HalfCloseableStream(Stream):
         """
 
 
-class Listener(AsyncResource):
+# A regular invariant generic type
+T = TypeVar("T")
+
+# The type of object produced by a ReceiveChannel (covariant because
+# ReceiveChannel[Derived] can be passed to someone expecting
+# ReceiveChannel[Base])
+ReceiveType = TypeVar("ReceiveType", covariant=True)
+
+# The type of object accepted by a SendChannel (contravariant because
+# SendChannel[Base] can be passed to someone expecting
+# SendChannel[Derived])
+SendType = TypeVar("SendType", contravariant=True)
+
+# The type of object produced by a Listener (covariant plus must be
+# an AsyncResource)
+T_resource = TypeVar("T_resource", bound=AsyncResource, covariant=True)
+
+
+class Listener(AsyncResource, Generic[T_resource]):
     """A standard interface for listening for incoming connections.
 
+    :class:`Listener` objects also implement the :class:`AsyncResource`
+    interface, so they can be closed by calling :meth:`~AsyncResource.aclose`
+    or using an ``async with`` block.
+
     """
     __slots__ = ()
 
@@ -499,22 +527,120 @@ class Listener(AsyncResource):
 
         Returns:
           AsyncResource: An object representing the incoming connection. In
-          practice this is almost always some variety of :class:`Stream`,
-          though in principle you could also use this interface with, say,
-          SOCK_SEQPACKET sockets or similar.
+          practice this is generally some kind of :class:`Stream`,
+          but in principle you could also define a :class:`Listener` that
+          returned, say, channel objects.
 
         Raises:
-          trio.ResourceBusyError: if two tasks attempt to call
+          trio.BusyResourceError: if two tasks attempt to call
               :meth:`accept` on the same listener at the same time.
           trio.ClosedResourceError: if you previously closed this listener
               object, or if another task closes this listener object while
               :meth:`accept` is running.
 
-        Note that there is no ``BrokenListenerError``, because for listeners
-        there is no general condition of "the network/remote peer broke the
-        connection" that can be handled in a generic way, like there is for
-        streams. Other errors *can* occur and be raised from :meth:`accept` –
-        for example, if you run out of file descriptors then you might get an
-        :class:`OSError` with its errno set to ``EMFILE``.
+        Listeners don't generally raise :exc:`~trio.BrokenResourceError`,
+        because for listeners there is no general condition of "the
+        network/remote peer broke the connection" that can be handled in a
+        generic way, like there is for streams. Other errors *can* occur and
+        be raised from :meth:`accept` – for example, if you run out of file
+        descriptors then you might get an :class:`OSError` with its errno set
+        to ``EMFILE``.
 
         """
+
+
+class SendChannel(AsyncResource, Generic[SendType]):
+    """A standard interface for sending Python objects to some receiver.
+
+    `SendChannel` objects also implement the `AsyncResource` interface, so
+    they can be closed by calling `~AsyncResource.aclose` or using an ``async
+    with`` block.
+
+    If you want to send raw bytes rather than Python objects, see
+    `SendStream`.
+
+    """
+    __slots__ = ()
+
+    @abstractmethod
+    async def send(self, value: SendType) -> None:
+        """Attempt to send an object through the channel, blocking if necessary.
+
+        Args:
+          value (object): The object to send.
+
+        Raises:
+          trio.BrokenResourceError: if something has gone wrong, and the
+              channel is broken. For example, you may get this if the receiver
+              has already been closed.
+          trio.ClosedResourceError: if you previously closed this
+              :class:`SendChannel` object, or if another task closes it while
+              :meth:`send` is running.
+          trio.BusyResourceError: some channels allow multiple tasks to call
+              `send` at the same time, but others don't. If you try to call
+              `send` simultaneously from multiple tasks on a channel that
+              doesn't support it, then you can get `~trio.BusyResourceError`.
+
+        """
+
+
+class ReceiveChannel(AsyncResource, Generic[ReceiveType]):
+    """A standard interface for receiving Python objects from some sender.
+
+    You can iterate over a :class:`ReceiveChannel` using an ``async for``
+    loop::
+
+       async for value in receive_channel:
+           ...
+
+    This is equivalent to calling :meth:`receive` repeatedly. The loop exits
+    without error when `receive` raises `~trio.EndOfChannel`.
+
+    `ReceiveChannel` objects also implement the `AsyncResource` interface, so
+    they can be closed by calling `~AsyncResource.aclose` or using an ``async
+    with`` block.
+
+    If you want to receive raw bytes rather than Python objects, see
+    `ReceiveStream`.
+
+    """
+    __slots__ = ()
+
+    @abstractmethod
+    async def receive(self) -> ReceiveType:
+        """Attempt to receive an incoming object, blocking if necessary.
+
+        Returns:
+          object: Whatever object was received.
+
+        Raises:
+          trio.EndOfChannel: if the sender has been closed cleanly, and no
+              more objects are coming. This is not an error condition.
+          trio.ClosedResourceError: if you previously closed this
+              :class:`ReceiveChannel` object.
+          trio.BrokenResourceError: if something has gone wrong, and the
+              channel is broken.
+          trio.BusyResourceError: some channels allow multiple tasks to call
+              `receive` at the same time, but others don't. If you try to call
+              `receive` simultaneously from multiple tasks on a channel that
+              doesn't support it, then you can get `~trio.BusyResourceError`.
+
+        """
+
+    def __aiter__(self):
+        return self
+
+    async def __anext__(self) -> ReceiveType:
+        try:
+            return await self.receive()
+        except trio.EndOfChannel:
+            raise StopAsyncIteration
+
+
+class Channel(SendChannel[T], ReceiveChannel[T]):
+    """A standard interface for interacting with bidirectional channels.
+
+    A `Channel` is an object that implements both the `SendChannel` and
+    `ReceiveChannel` interfaces, so you can both send and receive objects.
+
+    """
